@@ -24,8 +24,10 @@
 
 #include "bolt-enums.h"
 #include "bolt-error.h"
+#include "bolt-glue.h"
 #include "bolt-str.h"
 
+#include "test-enums.h"
 #include "bolt-test-resources.h"
 
 #include <glib.h>
@@ -34,16 +36,92 @@
 
 #include <locale.h>
 
+/* *** Tiny object with only an "id" property */
+#define BT_TYPE_ID bt_id_get_type ()
+G_DECLARE_FINAL_TYPE (BtId, bt_id, BT, ID, BoltExported);
+
+struct _BtId
+{
+  BoltExported parent;
+};
+
+G_DEFINE_TYPE (BtId, bt_id, BOLT_TYPE_EXPORTED);
+
+enum {
+  PROP_ID_0,
+  PROP_OBJECT_ID,
+  PROP_ID,
+  PROP_ID_LAST
+};
+
+
+static GParamSpec *id_props[PROP_ID_LAST] = {NULL, };
+
+
+static void
+bt_id_init (BtId *be)
+{
+}
+
+static void
+bt_id_get_property (GObject    *object,
+                    guint       prop_id,
+                    GValue     *value,
+                    GParamSpec *pspec)
+{
+  switch (prop_id)
+    {
+    case PROP_ID:
+    case PROP_OBJECT_ID:
+      g_value_set_string (value, "bolt-id");
+      break;
+
+    default:
+      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
+    }
+}
+
+static void
+bt_id_class_init (BtIdClass *klass)
+{
+  GObjectClass *gobject_class = G_OBJECT_CLASS (klass);
+
+  gobject_class->get_property = bt_id_get_property;
+
+  id_props[PROP_OBJECT_ID] =
+    bolt_param_spec_override (gobject_class,
+                              "object-id");
+
+  id_props[PROP_ID] =
+    g_param_spec_string ("id", "Id", NULL,
+                         NULL,
+                         G_PARAM_READABLE |
+                         G_PARAM_STATIC_NICK |
+                         G_PARAM_STATIC_BLURB);
+
+  g_object_class_install_properties (gobject_class,
+                                     PROP_ID_LAST,
+                                     id_props);
+
+
+}
 
 /*  */
 #define DBUS_IFACE "org.gnome.bolt.Example"
+#define DBUS_OPATH_BASE "/bolt/test"
 
 #define BT_TYPE_EXPORTED bt_exported_get_type ()
 G_DECLARE_FINAL_TYPE (BtExported, bt_exported, BT, EXPORTED, BoltExported);
 
-static gboolean  handle_ping (BoltExported          *obj,
-                              GVariant              *params,
-                              GDBusMethodInvocation *inv);
+static GVariant *  handle_ping (BoltExported          *obj,
+                                GVariant              *params,
+                                GDBusMethodInvocation *inv,
+                                GError               **error);
+
+static GVariant *  handle_peng (BoltExported          *obj,
+                                GVariant              *params,
+                                GDBusMethodInvocation *inv,
+                                GError               **error);
 
 static gboolean handle_authorize_method (BoltExported          *exported,
                                          GDBusMethodInvocation *inv,
@@ -67,19 +145,29 @@ static gboolean handle_set_security (BoltExported *obj,
                                      const GValue *value,
                                      GError      **error);
 
+static gboolean handle_set_kitt (BoltExported *obj,
+                                 const char   *name,
+                                 const GValue *value,
+                                 GError      **error);
+
 struct _BtExported
 {
-  BoltExported parent;
+  BoltExported  parent;
 
-  char        *str;
-  GError      *setter_err;
+  char         *object_id;
 
-  gboolean     prop_bool;
+  char         *str;
+  GError       *setter_err;
 
-  gboolean     authorize_methods;
-  gboolean     authorize_properties;
+  gboolean      prop_bool;
 
-  BoltSecurity security;
+  BtId         *prop_obj;
+
+  gboolean      authorize_methods;
+  gboolean      authorize_properties;
+
+  BoltSecurity  security;
+  BoltKittFlags kitt;
 };
 
 G_DEFINE_TYPE (BtExported, bt_exported, BOLT_TYPE_EXPORTED);
@@ -87,13 +175,17 @@ G_DEFINE_TYPE (BtExported, bt_exported, BOLT_TYPE_EXPORTED);
 enum {
   PROP_0,
 
+  PROP_MYID,
+
   PROP_STR,
   PROP_STR_RW,
   PROP_STR_RW_NOSETTER,
 
   PROP_BOOL,
+  PROP_OBJECT,
 
   PROP_SECURITY,
+  PROP_KITT,
 
   PROP_LAST
 };
@@ -105,7 +197,11 @@ bt_exported_finalize (GObject *object)
 {
   BtExported *be = BT_EXPORTED (object);
 
+  g_clear_object (&be->prop_obj);
+
   g_free (be->str);
+  g_free (be->object_id);
+
 
   G_OBJECT_CLASS (bt_exported_parent_class)->finalize (object);
 }
@@ -113,8 +209,9 @@ bt_exported_finalize (GObject *object)
 static void
 bt_exported_init (BtExported *be)
 {
+  be->object_id = g_strdup ("bt_exported0");
   be->str = g_strdup ("strfoo");
-
+  be->prop_obj = g_object_new (BT_TYPE_ID, NULL);
 }
 
 static void
@@ -127,6 +224,10 @@ bt_exported_get_property (GObject    *object,
 
   switch (prop_id)
     {
+    case PROP_MYID:
+      g_value_set_string (value, be->object_id);
+      break;
+
     case PROP_STR:
     case PROP_STR_RW:
     case PROP_STR_RW_NOSETTER:
@@ -137,11 +238,18 @@ bt_exported_get_property (GObject    *object,
       g_value_set_boolean (value, be->prop_bool);
       break;
 
+    case PROP_OBJECT:
+      g_value_set_object (value, be->prop_obj);
+      break;
 
     case PROP_SECURITY:
       g_value_set_enum (value, be->security);
       break;
 
+    case PROP_KITT:
+      g_value_set_flags (value, be->kitt);
+      break;
+
     default:
       G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
     }
@@ -172,6 +280,10 @@ bt_exported_set_property (GObject      *object,
       be->security = g_value_get_enum (value);
       break;
 
+    case PROP_KITT:
+      be->kitt = g_value_get_flags (value);
+      break;
+
     default:
       G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
     }
@@ -190,6 +302,11 @@ bt_exported_class_init (BtExportedClass *klass)
                                           DBUS_IFACE,
                                           "/bolt/tests/exported/example.bolt.xml");
 
+  bolt_exported_class_set_object_path (exported_class, DBUS_OPATH_BASE);
+
+  props[PROP_MYID] =
+    bolt_param_spec_override (gobject_class, "object-id");
+
   props[PROP_STR] =
     g_param_spec_string ("str", "StrFoo", NULL,
                          NULL,
@@ -218,6 +335,12 @@ bt_exported_class_init (BtExportedClass *klass)
                           G_PARAM_STATIC_NICK |
                           G_PARAM_STATIC_BLURB);
 
+  props[PROP_OBJECT] =
+    g_param_spec_object ("object", "Object", NULL,
+                         BT_TYPE_ID,
+                         G_PARAM_READABLE |
+                         G_PARAM_STATIC_NICK |
+                         G_PARAM_STATIC_BLURB);
 
   props[PROP_SECURITY] =
     g_param_spec_enum ("security", "Security", NULL,
@@ -227,12 +350,21 @@ bt_exported_class_init (BtExportedClass *klass)
                        G_PARAM_STATIC_NICK |
                        G_PARAM_STATIC_BLURB);
 
+  props[PROP_KITT] =
+    g_param_spec_flags ("kitt", "KittMode", NULL,
+                        BOLT_TYPE_KITT_FLAGS,
+                        BOLT_KITT_DISABLED,
+                        G_PARAM_READWRITE |
+                        G_PARAM_STATIC_NICK |
+                        G_PARAM_STATIC_BLURB);
+
   g_object_class_install_properties (gobject_class,
                                      PROP_LAST,
                                      props);
 
   bolt_exported_class_export_properties (exported_class, PROP_STR, PROP_LAST, props);
   bolt_exported_class_export_method (exported_class, "Ping", handle_ping);
+  bolt_exported_class_export_method (exported_class, "Peng", handle_peng);
 
   bolt_exported_class_property_setter (exported_class,
                                        props[PROP_STR_RW],
@@ -241,6 +373,10 @@ bt_exported_class_init (BtExportedClass *klass)
   bolt_exported_class_property_setter (exported_class,
                                        props[PROP_SECURITY],
                                        handle_set_security);
+
+  bolt_exported_class_property_setter (exported_class,
+                                       props[PROP_KITT],
+                                       handle_set_kitt);
 }
 
 static gboolean
@@ -299,17 +435,35 @@ bt_exported_install_property_authorizer (BtExported *be)
 }
 
 
-static gboolean
+static GVariant *
 handle_ping (BoltExported          *obj,
              GVariant              *params,
-             GDBusMethodInvocation *inv)
+             GDBusMethodInvocation *inv,
+             GError               **error)
 {
   //  BtExported *be = BT_EXPORTED (obj);
-
-  g_dbus_method_invocation_return_value (inv, g_variant_new ("(s)", "PONG"));
-  return TRUE;
+  return g_variant_new ("(s)", "PONG");
 }
 
+static GVariant *
+handle_peng (BoltExported          *obj,
+             GVariant              *params,
+             GDBusMethodInvocation *inv,
+             GError               **error)
+{
+  const char *str;
+
+  g_variant_get_child (params, 0, "&s", &str);
+  if (str == NULL)
+    str = "PENG";
+
+  g_set_error (error, BOLT_ERROR, BOLT_ERROR_FAILED,
+               "failing with: %s", str);
+
+  return NULL;
+}
+
+
 static gboolean
 handle_set_str_rw (BoltExported *obj,
                    const char   *name,
@@ -348,6 +502,17 @@ handle_set_security (BoltExported *obj,
   return TRUE;
 }
 
+static gboolean
+handle_set_kitt (BoltExported *obj,
+                 const char   *name,
+                 const GValue *value,
+                 GError      **error)
+{
+  BtExported *be = BT_EXPORTED (obj);
+
+  be->kitt = g_value_get_flags (value);
+  return TRUE;
+}
 
 /* *********** */
 
@@ -366,7 +531,9 @@ static void
 test_exported_setup (TestExported *tt, gconstpointer data)
 {
   g_autoptr(GError) err = NULL;
+  g_autofree char *opath = NULL;
   const char *obj_path = "/obj";
+  gboolean exported = FALSE;
   gboolean ok;
 
   tt->bus = g_bus_get_sync (G_BUS_TYPE_SESSION, NULL, &err);
@@ -390,6 +557,15 @@ test_exported_setup (TestExported *tt, gconstpointer data)
 
   g_assert_cmpstr (tt->obj_path, ==, obj_path);
   tt->bus_name = g_dbus_connection_get_unique_name (tt->bus);
+
+  g_object_get (tt->obj,
+                "object-path", &opath,
+                "exported", &exported,
+                NULL);
+
+  g_assert_cmpstr (obj_path, ==, opath);
+  g_assert_true (exported);
+  g_assert_true (bolt_exported_is_exported (BOLT_EXPORTED (tt->obj)));
 }
 
 static void
@@ -400,6 +576,7 @@ test_exported_teardown (TestExported *tt, gconstpointer data)
   ok = bolt_exported_unexport (BOLT_EXPORTED (tt->obj));
 
   g_assert_true (ok);
+  g_assert_false (bolt_exported_is_exported (BOLT_EXPORTED (tt->obj)));
 
   g_clear_object (&tt->bus);
   g_clear_object (&tt->obj);
@@ -428,15 +605,17 @@ call_ctx_reset (CallCtx *ctx)
   if (ctx->data)
     g_variant_unref (ctx->data);
 
-  if (ctx->error)
-    g_clear_error (&ctx->error);
+  g_clear_error (&ctx->error);
 }
 
 static void
 call_ctx_free (CallCtx *ctx)
 {
+  g_return_if_fail (ctx != NULL);
+
   g_main_loop_unref (ctx->loop);
   call_ctx_reset (ctx);
+  g_free (ctx);
 }
 
 G_DEFINE_AUTOPTR_CLEANUP_FUNC (CallCtx, call_ctx_free);
@@ -460,6 +639,73 @@ call_ctx_run (CallCtx *ctx)
   g_main_loop_run (ctx->loop);
 }
 
+/* the actual tests */
+static void
+test_exported_export (TestExported *unused, gconstpointer data)
+{
+  g_autoptr(GError) err = NULL;
+  g_autoptr(GDBusConnection) bus = NULL;
+  g_autoptr(BtExported) obj = NULL;
+  g_autofree char *want = NULL;
+  g_autofree char *have = NULL;
+  const char *obj_path = NULL;
+  gboolean exported = FALSE;
+  gboolean ok;
+
+  bus = g_bus_get_sync (G_BUS_TYPE_SESSION, NULL, &err);
+
+  g_assert_no_error (err);
+  g_assert_nonnull (bus);
+  g_assert_false (g_dbus_connection_is_closed (bus));
+
+  obj = g_object_new (BT_TYPE_EXPORTED, NULL);
+  g_assert_nonnull (obj);
+
+  /* we test auto object path generation */
+  want = g_build_path ("/", DBUS_OPATH_BASE, obj->object_id, NULL);
+  ok = bolt_exported_export (BOLT_EXPORTED (obj),
+                             bus,
+                             NULL,
+                             &err);
+
+  g_assert_no_error (err);
+  g_assert_true (ok);
+
+  obj_path = bolt_exported_get_object_path (BOLT_EXPORTED (obj));
+
+  g_assert_cmpstr (want, ==, obj_path);
+
+  g_object_get (obj,
+                "object-path", &have,
+                "exported", &exported,
+                NULL);
+
+  g_assert_cmpstr (want, ==, have);
+  g_assert_true (exported);
+  g_assert_true (bolt_exported_is_exported (BOLT_EXPORTED (obj)));
+
+  /* unexport */
+  ok = bolt_exported_unexport (BOLT_EXPORTED (obj));
+
+  g_assert_true (ok);
+  g_assert_false (bolt_exported_is_exported (BOLT_EXPORTED (obj)));
+
+  /* test we handle special chars in object id */
+  bolt_set_strdup (&obj->object_id, "object id-@$1");
+  ok = bolt_exported_export (BOLT_EXPORTED (obj),
+                             bus,
+                             NULL,
+                             &err);
+  g_assert_no_error (err);
+  g_assert_true (ok);
+
+  obj_path = bolt_exported_get_object_path (BOLT_EXPORTED (obj));
+
+  bolt_set_str (&want, g_build_path ("/", DBUS_OPATH_BASE, "object_id___1", NULL));
+  g_assert_cmpstr (want, ==, obj_path);
+}
+
+
 static void
 test_exported_basic (TestExported *tt, gconstpointer data)
 {
@@ -535,6 +781,22 @@ test_exported_basic (TestExported *tt, gconstpointer data)
   g_variant_get (ctx->data, "(&s)", &str);
 
   g_assert_cmpstr (str, ==, "PONG");
+
+  /* check error handling in BoltExported's method dispatching */
+  g_dbus_connection_call (bus,
+                          tt->bus_name,
+                          tt->obj_path,
+                          DBUS_IFACE,
+                          "Peng",
+                          g_variant_new ("(s)", "Out of cheese"),
+                          NULL,
+                          G_DBUS_CALL_FLAGS_NONE,
+                          2000,
+                          NULL,
+                          dbus_call_done,
+                          ctx);
+  call_ctx_run (ctx);
+  g_assert_error (ctx->error, BOLT_ERROR, BOLT_ERROR_FAILED);
 }
 
 static void
@@ -747,7 +1009,21 @@ props_changed_signal (GDBusConnection *connection,
 
   g_debug ("got prop changes signal");
   ctx->data = changed_properties; /* transfer ownership */
-  g_main_loop_quit (ctx->loop);
+  if (g_main_loop_is_running (ctx->loop))
+    g_main_loop_quit (ctx->loop);
+}
+
+static gboolean
+change_properties (gpointer user_data)
+{
+  TestExported *tt = user_data;
+
+  g_object_set (tt->obj,
+                "str-rw", "huhu",
+                "bool", TRUE,
+                NULL);
+
+  return G_SOURCE_REMOVE;
 }
 
 static void
@@ -776,17 +1052,11 @@ test_exported_props_changed (TestExported *tt, gconstpointer data)
 
   g_assert_cmpuint (sid, >, 0);
 
-  /* g_signal_connect (tt->obj, "notify::str-rw", */
-  /*                   G_CALLBACK (call_ctx_stop_on_notify), */
-  /*                   ctx); */
+  g_idle_add (change_properties, tt);
 
-  g_object_set (tt->obj,
-                "str-rw", "huhu",
-                "bool", TRUE,
-                NULL);
-
-  G_DEBUG_HERE ();
   call_ctx_run (ctx);
+  g_dbus_connection_signal_unsubscribe (tt->bus, sid);
+
   g_assert_no_error (ctx->error);
   g_assert_nonnull (ctx->data);
 
@@ -874,6 +1144,155 @@ test_exported_props_enums (TestExported *tt, gconstpointer data)
   g_assert_cmpint (tt->obj->security, ==, BOLT_SECURITY_USER);
 }
 
+static void
+test_exported_props_flags (TestExported *tt, gconstpointer data)
+{
+  g_autoptr(CallCtx) ctx = NULL;
+  g_autoptr(GVariant) v = NULL;
+  g_autofree char *want = NULL;
+  g_autofree char *ref = NULL;
+  const char *have = NULL;
+  BoltKittFlags kf = BOLT_KITT_ENABLED | BOLT_KITT_TURBO_BOOST;
+
+  ctx = call_ctx_new ();
+
+  tt->obj->kitt = BOLT_KITT_DEFAULT;
+
+  /* valid property value, should be converted to string */
+  g_dbus_connection_call (tt->bus,
+                          tt->bus_name,
+                          tt->obj_path,
+                          "org.freedesktop.DBus.Properties",
+                          "Get",
+                          g_variant_new ("(ss)",
+                                         DBUS_IFACE,
+                                         "KittMode"),
+                          G_VARIANT_TYPE ("(v)"),
+                          G_DBUS_CALL_FLAGS_NONE,
+                          2000,
+                          NULL,
+                          dbus_call_done,
+                          ctx);
+
+  call_ctx_run (ctx);
+  g_assert_no_error (ctx->error);
+
+  g_assert_nonnull (ctx->data);
+  g_variant_get (ctx->data, "(v)", &v);
+  have = g_variant_get_string (v, NULL);
+  want = bolt_flags_to_string (BOLT_TYPE_KITT_FLAGS,
+                               tt->obj->kitt,
+                               NULL);
+  g_assert_cmpstr (have, ==, want);
+
+  /* setter */
+  bt_exported_install_property_authorizer (tt->obj);
+  tt->obj->authorize_properties = TRUE;
+  ref = bolt_flags_to_string (BOLT_TYPE_KITT_FLAGS, kf, NULL);
+
+  g_dbus_connection_call (tt->bus,
+                          tt->bus_name,
+                          tt->obj_path,
+                          "org.freedesktop.DBus.Properties",
+                          "Set",
+                          g_variant_new ("(ssv)",
+                                         DBUS_IFACE,
+                                         "KittMode",
+                                         g_variant_new ("s", ref)),
+                          NULL,
+                          G_DBUS_CALL_FLAGS_NONE,
+                          2000,
+                          NULL,
+                          dbus_call_done,
+                          ctx);
+
+  call_ctx_run (ctx);
+  g_assert_no_error (ctx->error);
+  g_assert_cmpint (tt->obj->kitt, ==, kf);
+
+  /* setter with invalid argument */
+
+  g_dbus_connection_call (tt->bus,
+                          tt->bus_name,
+                          tt->obj_path,
+                          "org.freedesktop.DBus.Properties",
+                          "Set",
+                          g_variant_new ("(ssv)",
+                                         DBUS_IFACE,
+                                         "KittMode",
+                                         g_variant_new ("s", "invalid | foobar")),
+                          NULL,
+                          G_DBUS_CALL_FLAGS_NONE,
+                          2000,
+                          NULL,
+                          dbus_call_done,
+                          ctx);
+
+  call_ctx_run (ctx);
+  g_assert_error (ctx->error, G_DBUS_ERROR, G_DBUS_ERROR_INVALID_ARGS);
+}
+
+static void
+test_exported_props_object (TestExported *tt, gconstpointer data)
+{
+  g_autoptr(CallCtx) ctx = NULL;
+  g_autoptr(GVariant) v = NULL;
+  g_autofree char *want = NULL;
+  const char *have;
+
+  ctx = call_ctx_new ();
+
+  /* valid property value, should be converted to string */
+  g_dbus_connection_call (tt->bus,
+                          tt->bus_name,
+                          tt->obj_path,
+                          "org.freedesktop.DBus.Properties",
+                          "Get",
+                          g_variant_new ("(ss)",
+                                         DBUS_IFACE,
+                                         "Object"),
+                          G_VARIANT_TYPE ("(v)"),
+                          G_DBUS_CALL_FLAGS_NONE,
+                          2000,
+                          NULL,
+                          dbus_call_done,
+                          ctx);
+
+  call_ctx_run (ctx);
+  g_assert_no_error (ctx->error);
+
+  g_assert_nonnull (ctx->data);
+  g_variant_get (ctx->data, "(v)", &v);
+  have = g_variant_get_string (v, NULL);
+  g_object_get (tt->obj->prop_obj, "id", &want, NULL);
+
+  g_assert_cmpstr (have, ==, want);
+
+  /* setter, should be an error (invalid arguments) */
+  bt_exported_install_property_authorizer (tt->obj);
+  tt->obj->authorize_properties = TRUE;
+  have = bolt_security_to_string (BOLT_SECURITY_USER);
+
+  g_dbus_connection_call (tt->bus,
+                          tt->bus_name,
+                          tt->obj_path,
+                          "org.freedesktop.DBus.Properties",
+                          "Set",
+                          g_variant_new ("(ssv)",
+                                         DBUS_IFACE,
+                                         "Object",
+                                         g_variant_new ("s", have)),
+                          NULL,
+                          G_DBUS_CALL_FLAGS_NONE,
+                          2000,
+                          NULL,
+                          dbus_call_done,
+                          ctx);
+
+  call_ctx_run (ctx);
+  g_assert_error (ctx->error, G_DBUS_ERROR, G_DBUS_ERROR_INVALID_ARGS);
+}
+
 int
 main (int argc, char **argv)
 {
@@ -885,6 +1304,13 @@ main (int argc, char **argv)
 
   g_resources_register (bolt_test_get_resource ());
 
+  g_test_add ("/exported/export",
+              TestExported,
+              NULL,
+              NULL,
+              test_exported_export,
+              NULL);
+
   g_test_add ("/exported/basic",
               TestExported,
               NULL,
@@ -913,6 +1339,20 @@ main (int argc, char **argv)
               test_exported_props_enums,
               test_exported_teardown);
 
+  g_test_add ("/exported/props/flags",
+              TestExported,
+              NULL,
+              test_exported_setup,
+              test_exported_props_flags,
+              test_exported_teardown);
+
+  g_test_add ("/exported/props/object",
+              TestExported,
+              NULL,
+              test_exported_setup,
+              test_exported_props_object,
+              test_exported_teardown);
+
   test_bus = g_test_dbus_new (G_TEST_DBUS_NONE);
   g_test_dbus_up (test_bus);
   g_assert_nonnull (test_bus);
